/*
 *
 */


/*
 *
 */
package installer.common;


/*
 *
 */
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.List;
import java.util.logging.Logger;


/*
 * intended to run processes that print line-based output to stdout
 * (if they print anything at all)
 */
public class ProcessRunner
{
    /*
     *
     */
    public static class Result<OutputType>
    {
        public OutputType outputResult;
        public int processReturnResult;

        public Result(OutputType outputResult,
                      int processReturnResult)
        {
            this.outputResult = outputResult;
            this.processReturnResult = processReturnResult;
        }
    }


    /*
     *
     */
    private static final int PROCESS_TIMEOUT_MILLIS = 120000;


    /*
     *
     */
    private static <OutputType> Thread
        createProcessRunnerThread(final char[] stdinCharArray,
                                  final File workingDirectory,
                                  final ProcessOutputAnalyzer<OutputType> outputAnalyzer,
                                  final Pointer<Result<OutputType>> resultPointer,
                                  final Pointer<ProcessExecutionException> exceptionPointer,
                                  final String... commandLine)
    {
        return new Thread() {
            public void run()
            {
                Process process;
                BufferedReader stdoutReader;
                OutputType outputAnalysisResult;
                int processReturnValue;
                Logger logger;
                
                
                //
                logger = Logger.getLogger("controller");

                //
                try
                {
                    // FIXME: should escape double quotes in the strings, so they look nicer
                    {
                        String cmdString = "";
                        for (String arg : commandLine)
                        {
                            cmdString += arg + " ";
                        }
                        logger.finer("About to execute " + cmdString +
                                     ((workingDirectory != null)
                                      ? ("with working dir. " + workingDirectory)
                                      : "")
                                     + "...\n");
                    }

                    process = Runtime.getRuntime().exec(commandLine, null, workingDirectory);
                }
                catch (IOException ex)
                {
                    exceptionPointer.object = 
                        new ProcessExecutionException(commandLine, ex.getMessage());
                    return;
                }

                //
                if (stdinCharArray != null)
                {
                    OutputStreamWriter stdinWriter =
                        new OutputStreamWriter(process.getOutputStream());

                    try
                    {
                        stdinWriter.write(stdinCharArray, 0, stdinCharArray.length);
                        stdinWriter.close();
                    }
                    catch (IOException ex)
                    {
                        exceptionPointer.object =
                            new ProcessExecutionException(commandLine, ex.getMessage());
                        return;
                    }
                }

                //
                stdoutReader = new BufferedReader(new InputStreamReader(process.getInputStream()));

                //
                try
                {
                    if (outputAnalyzer != null)
                    {
                        outputAnalysisResult = outputAnalyzer.analyzeOutput(stdoutReader);
                    }
                    else
                    {
                        outputAnalysisResult = null;
                    }
                }
                catch (IOException ex)
                {
                    exceptionPointer.object =
                        new ProcessExecutionException(commandLine, ex.getMessage());
                    return;
                }
                catch (ProcessOutputAnalyzer.UnexpectedOutputException ex)
                {
                    exceptionPointer.object =
                        new ProcessExecutionException(commandLine, ex.getMessage());
                    return;
                }
                finally
                {
                    try
                    {
                        stdoutReader.close();
                    }
                    catch (IOException ex)
                    {
                        // FIXME: this is not an important exception, but it should be logged
                    }

                    try
                    {
                        processReturnValue = process.waitFor();
                        logger.finer("Process " + commandLine[0] + " finished running and returned "
                                     + processReturnValue + ".\n");
                    }
                    catch (InterruptedException ex)
                    {
                        process.destroy();
                        // this exception will likely be ignored
                        exceptionPointer.object =
                            new ProcessExecutionException(commandLine, ex.getMessage());
                        return;
                    }
                }

                //
                resultPointer.object =
                    new Result<OutputType>(outputAnalysisResult, processReturnValue);
            }
        };
    }


    /*
     *
     */
    public static <OutputType> Result<OutputType>
        run(ProcessOutputAnalyzer<OutputType> outputAnalyzer,
            String... commandLine)
        throws ProcessExecutionException
    {
        return runWithWorkingDirectory(outputAnalyzer, null, commandLine);
    }


    /*
     * outputAnalyzer may be null, in which case
     * no output analysis is done;
     *
     * stdinCharArray may be null, but if it is not,
     * it is used to write into stdin of the process,
     * before anything is read from stdout; it is not
     * modified
     *
     * commandLine should have at least one element,
     * and all of its elements should be non-null;
     * if processReturnValue is non-null, it is filled
     * with the value returned by the process
     *
     * FIXMEFIXME: build in an optional timer!!!
     *
     * FIXME: probably should not convert all IOException's to
     *        ProcessExecutionException's
     */
    public static <OutputType> Result<OutputType>
        run(char[] stdinCharArray,
            ProcessOutputAnalyzer<OutputType> outputAnalyzer,
            String... commandLine)                            
        throws ProcessExecutionException
    {
        return runWithWorkingDirectoryAndTimeout(stdinCharArray, null, outputAnalyzer, commandLine);
    }


    /*
     *
     */
    public static <OutputType> Result<OutputType>
        runWithWorkingDirectory(ProcessOutputAnalyzer<OutputType> outputAnalyzer,
                                File workingDirectory,
                                String... commandLine)
        throws ProcessExecutionException
    {
        return runWithWorkingDirectoryAndTimeout(null,
                                                 workingDirectory,
                                                 outputAnalyzer,
                                                 commandLine);
    }
    

    /*
     * stdinCharArray is not changed
     *
     * FIXME: it would probably be better to throw up a dialog asking whether to re-run
     *        with a longer timeout
     */
    private static <OutputType> Result<OutputType>
        runWithWorkingDirectoryAndTimeout(final char[] stdinCharArray,
                                          final File workingDirectory,
                                          final ProcessOutputAnalyzer<OutputType> outputAnalyzer,
                                          final String... commandLine)
        throws ProcessExecutionException
    {
        Thread procThread;
        final Pointer<Result<OutputType>> resultPointer;
        final Pointer<ProcessExecutionException> exceptionPointer;


        //
        resultPointer = new Pointer<Result<OutputType>>(null);
        exceptionPointer = new Pointer<ProcessExecutionException>(null);

        //
        procThread = createProcessRunnerThread(stdinCharArray,
                                               workingDirectory,
                                               outputAnalyzer,
                                               resultPointer,
                                               exceptionPointer,
                                               commandLine);
        
        //
        procThread.start();
        try {
            procThread.join(PROCESS_TIMEOUT_MILLIS);
        } catch (InterruptedException ex) {
            // this should never happen
        }
        if (procThread.getState() == Thread.State.TERMINATED)
        {
            if (exceptionPointer.object != null)
            {
                throw exceptionPointer.object;
            }
            else
            {
                return resultPointer.object;
            }
        }
        else
        {
            procThread.interrupt();
            while (procThread.getState() != Thread.State.TERMINATED)
            {
                try {
                    procThread.join();
                } catch (InterruptedException ex) {
                    // this should never happen, but we still loop
                    // to kill the thread
                }
            }
            throw new ProcessExecutionException(commandLine,
                                                "Timed out after "
                                                + (PROCESS_TIMEOUT_MILLIS / 1000.0) + " seconds");
        }
    }
}