/*
 *
 */


/*
 *
 */
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Logger;

import installer.common.Controller;
import installer.common.InstallerGUI;
import installer.common.NoSuperuserPasswordException;
import installer.common.ProcessExecutionException;
import installer.common.ProcessOutputAnalyzer;
import installer.common.ProcessRunner;
import installer.common.ResourceNotFoundException;


/*
 *
 */
class NativeController extends Controller
{
    /*
     *
     */
    static final String nativeI386BinaryListFileName = "native-lib-files-i386.txt";
    static final String nativeAMD64BinaryListFileName = "native-lib-files-amd64.txt";


    /*
     *
     */
    static final String withttyAMD64NameFileName = "withtty-amd64-name.txt";
    static final String withttyI386NameFileName = "withtty-i386-name.txt";


    // /*
    //  *
    //  */
    // private Logger logger;


    /*
     *
     */
    private String suPath;
    private String sudoPath;


    /*
     *
     */
    private String withttyName;
    private File withttyPath;

    
    /*
     *
     */
    private boolean nativeSpecificFilesCopied;


    /*
     *
     */
    public NativeController()
    {
        this.suPath = null;
        this.sudoPath = null;
        this.withttyName = null;
        this.withttyPath = null;
        this.nativeSpecificFilesCopied = false;
    }


    /*
     * FIXME: should do a "which cp" to make sure it's /bin/cp
     */
    protected void copyFile(File sourceFile,
                            File destinationFile)
        throws IOException,
               NoSuperuserPasswordException,
               ProcessExecutionException
    {
        ProcessRunner.Result<?> cpResult;
        String[] commandLine =
            { "/bin/cp", sourceFile.getAbsolutePath(), destinationFile.getAbsolutePath() };

        
        //
        cpResult = ProcessRunner.run(null, commandLine);

        //
        if (cpResult.processReturnResult != 0)
        {
            //
            this.logger.info("Tried to copy " + sourceFile.getAbsolutePath() + " to "
                             + destinationFile.getAbsolutePath() + ", but failed.\n");
            
            // FIXME: we're assuming the failure was because we aren't superuser;
            //        this is not exactly robust, it's just the most likely outcome
            ensureSuperuserPasswordAvailable();


            //
            cpResult = runAsRoot(null, commandLine);

            //
            if (cpResult.processReturnResult != 0)
            {
                // FIXME: this should not really be an IOException...
                throw new IOException("Unable to copy "
                                      + sourceFile.getAbsolutePath()
                                      + " to " + destinationFile.getAbsolutePath()
                                      + ", even as superuser (command return value was "
                                      + cpResult.processReturnResult + ")");
            }
            else
            {
                this.logger.info("Copied " + sourceFile.getAbsolutePath() + " to "
                                 + destinationFile.getAbsolutePath() + " successfully.\n");
            }
        }
        else
        {
            this.logger.info("Copied " + sourceFile.getAbsolutePath() + " to "
                             + destinationFile.getAbsolutePath() + " successfully.\n");
        }
    }


    /*
     * withtty is a helper program that runs a command in a pseudo-tty (necessary for su
     * and sometimes sudo)
     *
     * arch should be an available arch that can run withtty
     * (should be either "i386" or "amd64" at the moment)
     */
    protected void copyNativeSpecificFilesFromPackageToHost(String arch)
        throws IOException,
               ProcessExecutionException,
               ResourceNotFoundException
    {
        Set<String> withttyNames;

        
        // we never need to do this twice!
        if (this.nativeSpecificFilesCopied)
        {
            return;
        }

        //
        if (arch.equals("i386"))
        {
            withttyNames = getResourceNamesFromTextFileResource(withttyI386NameFileName);
        }
        else if (arch.equals("amd64"))
        {
            withttyNames = getResourceNamesFromTextFileResource(withttyAMD64NameFileName);
        }
        else
        {
            // this should never happen; if it does, there's a bug
            // FIXME: perhaps slightly improper use of IOException
            throw new IOException("unavailable arch. requested for withtty: " + arch);
        }
        assert(withttyNames.size() == 1);

        // FIXME: need to perform a cleanup on withtty
        this.withttyName = withttyNames.iterator().next();
        this.withttyPath = new File(getActiveConfigDirectory(),
                                    localizeRelativePath(this.withttyName));
        copyResourceToLocalFile(this.withttyName, this.withttyPath);
        makeFileExecutable(this.withttyPath);

        //
        this.nativeSpecificFilesCopied = true;
    }


    /*
     *
     */
    private Set<File> findMatlabPaths()
        throws ProcessExecutionException
    {
        String matlabPath;
        Set<File> matlabDirs;
        ProcessOutputAnalyzer<String> outputAnalyzer;
        ProcessRunner.Result<String> matlabPathResult;


        //
        outputAnalyzer =
            new ProcessOutputAnalyzer.FirstLineOnly<String>(new ProcessOutputAnalyzer.LastLine());
        matlabPathResult =
            ProcessRunner.run(outputAnalyzer,
                              "bash",
                              "-c",
                              "matlab -e | sed '/^MATLAB=\\(.*\\)$/s/^MATLAB=\\(.*\\)$/\\1/p;d'");
        
        //
        matlabDirs = new HashSet<File>();
        matlabDirs.add(new File(matlabPathResult.outputResult));

        //
        return matlabDirs;
    }


    /*
     *
     */
    public Map<File, InstallerGUI.MatlabInstallation> findMatlabPathsAndVersions()
    {
        Set<File> matlabPaths;
        Map<File, InstallerGUI.MatlabInstallation> matlabPathsToVersions;


        //
        matlabPathsToVersions = new HashMap<File, InstallerGUI.MatlabInstallation>();

        //
        try
        {
            this.logger.info("Looking for MATLAB paths...\n");
            matlabPaths = findMatlabPaths();
            this.logger.info("Done looking for MATLAB paths.\n");
        }
        catch (ProcessExecutionException ex)
        {
            this.logger.warning("Unable to find MATLAB paths: " + ex.getMessage() + "\n");

            // matlabPathsToVersions is empty here
            return matlabPathsToVersions;
        }
        
        //
        for (File matlabPath : matlabPaths)
        {
            InstallerGUI.MatlabInstallation matlabInst =
                verifyMatlabPathAndGetVersion(matlabPath);

            if (matlabInst == null)
            {
                this.logger.severe("Unable to verify MATLAB installation in "
                                   + matlabPath.getAbsolutePath() + "\n");
                continue;
            }
            else
            {
                matlabPathsToVersions.put(matlabPath, matlabInst);
            }
        }

        //
        return matlabPathsToVersions;
    }


    // /*
    //  *
    //  */
    // private String findMatlabVersionFromPath(File matlabPath)
    //     throws FileNotFoundException,   // subclass of IOException
    //            IOException,
    //            ProcessExecutionException
    // {
    //     BufferedReader lineReader;
    //     String versionLine;


    //     //
    //     lineReader = new BufferedReader(new FileReader(new File(matlabPath, ".VERSION")));

    //     // FIXME FINISHME!!!
    //     return null;
    // }


    /*
     *
     */
    protected boolean
        gatherPlatformDependentConfigurationInformation(List<String> errors) // out
    {
        boolean fullySucceeded;


        //
        this.suPath = null;
        this.sudoPath = null;
        fullySucceeded = true;

        //
        try
        {
            this.suPath =
                ProcessRunner.run(new ProcessOutputAnalyzer.LastLine(), "which", "su").outputResult;
        }
        catch (ProcessExecutionException ex)
        {
            fullySucceeded = false;
            errors.add("Unable to get location of su: " + ex.getMessage());
        }

        try
        {
            this.sudoPath =
                ProcessRunner.run(new ProcessOutputAnalyzer.LastLine(),
                                  "which", "sudo").outputResult;
        }
        catch (ProcessExecutionException ex)
        {
            fullySucceeded = false;
            errors.add("Unable to get location of sudo: " + ex.getMessage());
        }

        //
        return fullySucceeded;
    }

    
    /*
     *
     */
    protected File getMatlabUserpathGuess()
    {
        return new File(getUserHomeDirectory(), "matlab");
    }


    /*
     *
     */
    protected Set<String> getNativeBinaryFileNames(String arch)
        throws IOException
    {
        if (arch.equals("i386"))
        {
            return getResourceNamesFromTextFileResource(nativeI386BinaryListFileName);
        }
        else if (arch.equals("amd64"))
        {
            return getResourceNamesFromTextFileResource(nativeAMD64BinaryListFileName);
        }
        else
        {
            // this should never happen
            throw new IOException("Unsupported arch. libs. requested (this is a bug!)");
        }
    }


    // /*
    //  *
    //  */
    // private Set<String> getNativeTempFileNames()
    //     throws IOException
    // {
    //     return getResourceNamesFromTextFileResource(nativeTempListFileName);
    // }


    /*
     * FIXME FILLME: currently unimplemented, because only critical on windows!
     */
    protected boolean isFileReadOnly(File file)
        throws IOException
    {
        return false;
    }


    /*
     * arch should be non-null
     */
    protected boolean isSupportedArch(String arch)
    {
        return arch.equals("i386") || arch.equals("amd64");
    }


    /*
     * version should be non-null
     *
     * FIXME FILLME
     */
    protected boolean isSupportedMatlabVersion(String version)
    {
        return true;
    }


    /*
     *
     */
    protected void launchMatlabWithCommand(File matlabPath,
                                           String matlabCommand,
                                           File workingDirectory)
        throws IOException
    {
        LinuxMatlabRunner.get().launch(matlabPath, matlabCommand, workingDirectory);
    }


    /*
     *
     */
    private void makeFileExecutable(File exeFile)
        throws ProcessExecutionException
    {
        ProcessRunner.Result<?> chmodResult;
        String[] commandLine;


        //
        commandLine = new String[] { "chmod", "+x", exeFile.getAbsolutePath() };
        chmodResult = ProcessRunner.run(null, commandLine);
        if (chmodResult.processReturnResult != 0)
        {
            throw new ProcessExecutionException(commandLine, "chmod didn't return 0");
        }
    }


    /*
     * originalCommandLine should be non-null, should contain at least
     * one element, and each of its elements should be non-null
     *
     * FIXME: should be able to use sudo or su
     */
    private <OutputType> ProcessRunner.Result<OutputType>
        runAsRoot(ProcessOutputAnalyzer<OutputType> outputAnalyzer,
                  String... originalCommandLine)
        throws ProcessExecutionException
    {
        String[] sudoCommandLine = null; // FIXME: rename to commandLine
        int i;
        
        
        // FIXME: this is a bit of a hack
        if (!this.withttyPath.exists())
        {
            withttyPath = new File(getActiveConfigDirectory(), this.withttyName);
        }
        
        //
        switch (getSuperuserPassword().authenticationMethod)
        {
        case SU:
            // build the root command line -- not as nasty as it looks;
            // e.g., {"cp", "X", "Y"} becomes
            //       {"withtty", "su", "DUMMY_PROMPT", "/bin/sh", "-c",
            //        "su -s /bin/sh -c \"$1 $2 $3\"",
            //        "su", "cp", "X", "Y"}
            sudoCommandLine = new String[7 + originalCommandLine.length];
            sudoCommandLine[0] = this.withttyPath.getAbsolutePath();
            sudoCommandLine[1] = "su";
            sudoCommandLine[2] = "DUMMY_PROMPT";
            sudoCommandLine[3] = "/bin/sh";
            sudoCommandLine[4] = "-c";
            sudoCommandLine[5] = this.suPath + " -s /bin/sh -c \"";
            for (i = 1; i <= originalCommandLine.length; ++i)
            {
                sudoCommandLine[5] += " $" + i;
            }
            sudoCommandLine[5] += "\"";
            sudoCommandLine[6] = "su";
            for (i = 0; i < originalCommandLine.length; ++i)
            {
                sudoCommandLine[7 + i] = originalCommandLine[i];
            }
            break;
            
        case SUDO:
            // build the root command line -- not as nasty as it looks;
            // e.g., {"cp", "X", "Y"} becomes
            //       {"withtty", "sudo", "WSM_INSTALL_SUDO_PROMPT", "/bin/sh", "-c",
            //        "sudo -K; sudo -p WSM_INSTALL_SUDO_PROMPT $1 $2 $3",
            //        "sudo", "cp", "X", "Y"}
            sudoCommandLine = new String[7 + originalCommandLine.length];
            sudoCommandLine[0] = this.withttyPath.getAbsolutePath();
            sudoCommandLine[1] = "sudo";
            sudoCommandLine[2] = "WSM_INSTALL_SUDO_PROMPT";
            sudoCommandLine[3] = "/bin/sh";
            sudoCommandLine[4] = "-c";
            sudoCommandLine[5] = this.sudoPath + " -K; "
                + this.sudoPath + " -p WSM_INSTALL_SUDO_PROMPT";
            for (i = 1; i <= originalCommandLine.length; ++i)
            {
                sudoCommandLine[5] += " $" + i;
            }
            sudoCommandLine[6] = "sudo";
            for (i = 0; i < originalCommandLine.length; ++i)
            {
                sudoCommandLine[7 + i] = originalCommandLine[i];
            }
            break;
        }
        
        //
        return ProcessRunner.run(getSuperuserPassword().passwordText,
                                 outputAnalyzer, sudoCommandLine);
    }


    /*
     *
     */
    public <ResultType> ResultType
        runMatlabCommand(File matlabPath,
                         String matlabCommand,
                         File logFile,
                         ProcessOutputAnalyzer<ResultType> outputAnalyzer)
        throws IOException,
               ProcessExecutionException
    {
        return LinuxMatlabRunner.get().run(matlabPath, matlabCommand, logFile, outputAnalyzer);
    }


    /*
     * FIXME FILLME: currently unimplemented because only critical on windows
     */
    protected void setFileReadOnly(File file,
                                   boolean readOnly)
        throws IOException
    {
    }


    // /*
    //  * FIXME: this is a bit brittle...
    //  */
    // public void setLogger(Logger logger)
    // {
    //     this.logger = logger;
    // }


    // /*
    //  * FIXME: this makes two calls to matlab, which is a bit silly
    //  * FIXME: also, the arch. check is hard-coded -- this is just temporary
    //  */
    // public InstallerGUI.MatlabInstallation verifyMatlabPathAndGetVersion(File matlabPath)
    //     throws IOException
    // {
    //     final String matlabVersionAndArchCommand =
    //         "nl=['\\','n'];fprintf(1,['%s',nl,'%s',nl], "
    //         + "version, "
    //         + "char(java.lang.System.getProperty('os.arch')));"
    //         + "exit;";


    //     //
    //     try
    //     {
    //         String arch;
    //         String version;
    //         List<String> fullOutput;

    //         //
    //         fullOutput = runMatlabCommand(matlabPath, matlabVersionAndArchCommand, null,
    //                                       new ProcessOutputAnalyzer.EveryLine());
    //         if (fullOutput == null
    //             || (fullOutput.size() < 2))
    //         {
    //             logger.severe("Unable to get version and/or arch. for MATLAB in "
    //                           + matlabPath.getAbsolutePath() + "\n");
    //             return null;
    //         }

    //         //
    //         version = fullOutput.get(fullOutput.size()-2);
    //         arch = fullOutput.get(fullOutput.size()-1);

    //         //
    //         if (arch == null
    //             || (!arch.equals("amd64") && !arch.equals("i386")))
    //         {
    //             logger.severe("Unable to use MATLAB in " + matlabPath.getAbsolutePath() + " -- "
    //                           + "the JVM's architecture is "
    //                           + ((arch == null) ? "unknown (Java property os.arch missing!)"
    //                              : arch) + ", "
    //                           + "but currently only amd64 and i386 are supported!\n");
    //             return null;
    //         }

    //         // FIXME: verify version more completely

    //         //
    //         return new InstallerGUI.MatlabInstallation(version, arch);
    //     }
    //     catch (IOException ex)
    //     {
    //         this.logger.warning("Unable to verify MATLAB installation in " + matlabPath + ": "
    //                             + ex.getMessage() + "\n");
    //         return null;
    //     }
    //     catch (ProcessExecutionException ex)
    //     {
    //         this.logger.warning("Unable to verify MATLAB installation in " + matlabPath + ": "
    //                             + ex.getMessage() + "\n");
    //         return null;
    //     }
    // }


    /*
     * return true if the superuser password is correct; false if not
     */
    protected boolean verifySuperuserPassword(Password superuserPassword)
    {
        String[] commandLine;
        ProcessRunner.Result<?> processResult;


        //
        switch (superuserPassword.authenticationMethod)
        {
        case SU:
            // FIXME: escape suPath
            // this looks like: withtty su DUMMY_PROMPT su -s /bin/sh -c exit
            commandLine =
                new String[] { this.withttyPath.getAbsolutePath(), "su", "DUMMY_PROMPT",
                               this.suPath, "-s", "/bin/sh", "-c", "exit" };
            
            try
            {
                processResult = ProcessRunner.run(superuserPassword.passwordText,
                                                  null,
                                                  commandLine);
            }
            catch (ProcessExecutionException ex)
            {
                this.logger.severe("su (under withtty) failed to run properly: "
                                   + ex.getMessage() + "\n");
                return false;
            }

            return (processResult.processReturnResult == 0);

        case SUDO:
            // FIXME: escape sudoPath
            // this looks like:
            //   withtty sudo WSM_INSTALL_SUDO_PROMPT /bin/sh -c "sudo -K; sudo /bin/sh -c exit"
            commandLine =
                new String[] { this.withttyPath.getAbsolutePath(), "sudo", "WSM_INSTALL_SUDO_PROMPT",
                               "/bin/sh", "-c",
                               this.sudoPath + " -K; "
                               + this.sudoPath + " -p WSM_INSTALL_SUDO_PROMPT /bin/sh -c exit" };

            // commandLine =
            //     new String[] { this.withttyPath.getAbsolutePath(),
            //     "/bin/sh", "-c", "exit" };

            try
            {
                processResult = ProcessRunner.run(superuserPassword.passwordText,
                                                  null,
                                                  commandLine);
            }
            catch (ProcessExecutionException ex)
            {
                this.logger.severe("sudo (under withtty) failed to run properly: "
                                   + ex.getMessage() + "\n");
                return false;
            }
            
            return (processResult.processReturnResult == 0);
        }

        // java should be able to figure out that this isn't necessary...
        return false;
    }
}

